#include <QSettings>
#include <QItemSelection>
#include <QGraphicsScene>
#include <QGraphicsPixmapItem>
#include <QGraphicsRectItem>
#include <QPixmap>
#include <QPropertyAnimation>
#include <QGraphicsBlurEffect>
#include "musiclibrarygui.h"
#include "ui_musiclibrarygui.h"
#include "musiclibrarybuilder.h"
#include "musiclibrarymodel.h"
#include "animatedpixmap.h"

MusicLibraryGUI::MusicLibraryGUI(QWidget *parent) :
    QMainWindow(parent), ui(new Ui::MusicLibraryGUI), forward(true)
{
    // Setup UI
    ui->setupUi(this);

    // Build a music library and create a model out of it
    library = MusicLibraryBuilder::build(this);
    model = new MusicLibraryModel(library);

    // Initialize the views with the model
    ui->treeView->setModel(model->tree());
    ui->listView->setModel(model->artists());

    // Add tree view selection handlers to display properties in text editors
    connect(ui->treeView->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)),
            SLOT(treeSelectionChanged(QItemSelection,QItemSelection)));

    // Setup graphics scene
    ui->graphicsView->setScene(new QGraphicsScene(QRectF(0, 0, 300, 120), this));
    setupScene();

    // Load UI configuration from settings
    QSettings settings("Symbio", "MusicLibraryGUI");
    ui->comboBox->setCurrentIndex(settings.value("Combo", 0).toInt());

    // Start a timer, which updates the graphics scene
    startTimer(2000);
}

MusicLibraryGUI::~MusicLibraryGUI()
{
    // Save configuration to settings on exit
    QSettings settings("Symbio", "MusicLibraryGUI");
    settings.setValue("Combo", ui->comboBox->currentIndex());
    delete ui;
}

void MusicLibraryGUI::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MusicLibraryGUI::listChanged(int index)
{
    // When something is selected from the combo box,
    // the list model within the list view is changed
    switch(index) {
    case 0:
        ui->listView->setModel(model->artists());
        break;
    case 1:
        ui->listView->setModel(model->records());
        break;
    case 2:
        ui->listView->setModel(model->songs());
        break;
    }
}

void MusicLibraryGUI::treeSelectionChanged(const QItemSelection &selected, const QItemSelection &)
{
    // When selection changes, the line text editors are updated with the contents
    // of the MusicObject that was selected.
    // The item selection object contains a list of model indexes that were selected
    QModelIndexList list = selected.indexes();
    if (list.size() == 1) {

        // If the selection list contains exactly one element, this gets the
        // model index from the list and fetches the MusicObject that is
        // represented by that particular index
        QModelIndex index = list.at(0);
        MusicObject *obj = static_cast<MusicObject *>(qVariantValue<QObject *>(index.data(Qt::UserRole)));

        // The name is a common property of all music objects, so it's available
        // via the MusicObject base class
        ui->nameEdit->setText(obj->name());

        // Other properties are object-specific, so this needs to determine
        // which kind of object was selected.
        // qobject_cast<> returns the pointer value if object is of that type.
        // If not, it returns a 0-value pointer
        Artist *a = qobject_cast<Artist *>(obj);
        if (a) {
            // If object was an Artist, display the home page in text editor
            ui->dataEdit->setText(a->homePage().toString());
        } else {
            Record *r = qobject_cast<Record *>(obj);
            if (r) {
                // If object was a Record, display the release date
                // Now this is using the text edit, but should really use
                // a date editor widget
                ui->dataEdit->setText(r->releaseDate().toString("ddd MMM d yyyy"));
            } else {
                // If not Artist or Record, it's a Song. Thus, display the song number.
                // This is also using the text editor, but should use a spinbox
                ui->dataEdit->setText(QString::number(static_cast<Song *>(obj)->number()));
            }
        }
    }
}

void MusicLibraryGUI::updateClicked()
{
    // Current selection can be fetched from the view selection model.
    // The MusicObject if fetched similarly as in selectionChanged function.
    // Probably would be a good idea to add a common function for handling
    // QModelIndexList to MusicObject mapping
    QModelIndexList list = ui->treeView->selectionModel()->selectedIndexes();
    if (list.size() == 1) {
        QModelIndex index = list.at(0);
        MusicObject *obj = static_cast<MusicObject *>(qVariantValue<QObject *>(index.data(Qt::UserRole)));

        // In this case the properties are updated, so the set-functions are used.
        // Name is common property and other properties need the qobject_cast
        obj->setName(ui->nameEdit->text());
        Artist *a = qobject_cast<Artist *>(obj);
        if (a) {
            a->setHomePage(QUrl(ui->dataEdit->text()));
        } else {
            Record *r = qobject_cast<Record *>(obj);
            if (r) {
                r->setReleaseDate(QDateTime::fromString(ui->dataEdit->text(), "ddd MMM d yyyy"));
            } else {
                static_cast<Song *>(obj)->setNumber(ui->dataEdit->text().toInt());
            }
        }
    }
}

void MusicLibraryGUI::setupScene()
{
    // Graphics view is used to display the cover images of each record
    QList<Record *> records = library->records();

    // currentIndex identifies the image that's shown on front of other images
    currentIndex = records.count() / 2;

    // For each record found, create a QPixmap from the file name stored into record.
    // The pixmap is scaled and wrapped into a AnimatedPixmap class, which derives
    // from QObject and QGraphicsPixmapItem
    //  - QObject enables property animations
    //  - QGraphicsPixmapItem allows it to be integrated to the graphics scene
    foreach (Record *record, records) {
        QPixmap pixmap(record->coverArtFile());
        pixmap = pixmap.scaled(100, 100, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
        AnimatedPixmap *coverImage = new AnimatedPixmap(pixmap);
        ui->graphicsView->scene()->addItem(coverImage);

        // Each pixmap item is stored into array, so their positions can
        // later be changed
        coverImages.append(coverImage);
    }

    // Updates the scene
    updateScene();
}

void MusicLibraryGUI::updateScene()
{
    // Find the center of scene
    int centerX = ui->graphicsView->sceneRect().width() / 2 - 50;
    int centerY = ui->graphicsView->sceneRect().height() / 2 - 50;

    // Margin is the amount of pixels between images
    int margin = 60;
    int index = 0;

    // Iterates over the record cover page pixmaps and sets a new position for them
    // The position changes are animated with the help of a property animation,
    // which moves the pixmap from current position to new position
    foreach (AnimatedPixmap *coverImage, coverImages) {

        // The animation start value needs to be updated depending on the
        // direction we're moving. Animation applies a translation to the
        // image, starting from the margin value and ending at 0. Thus,
        // since the images move the amount specified by margin, the
        // translation with margin amount of pixels means that the animation
        // starts from the current position
        coverImage->anim()->setStartValue(forward ? 60 : -60);
        coverImage->anim()->start();

        // Each record is positioned according to difference from current index
        int diff = currentIndex - index;
        coverImage->setPos(centerX - diff * margin, centerY);

        // The image at currentIndex list position has diff 0, and thus
        // is the "active" cover page. It's z order is 0 and others
        // get negative values, so they're shown behind the active one.
        coverImage->setZValue(diff < 0 ? diff : 0 - diff);
        index++;
    }
}

void MusicLibraryGUI::timerEvent(QTimerEvent *)
{
    if (forward) {
        // If moving forward, select next cover image as the "active" one
        currentIndex++;
    } else {
        // If moving backward, select previous
        currentIndex--;
    }
    if (currentIndex == library->records().size()) {
        // When end of list is reached, jump to previous and continue backwards
        currentIndex -= 2;
        forward = false;
    } else if (currentIndex == -1) {
        // When beginning of list is reached, jump the next and continue forwards
        currentIndex += 2;
        forward = true;
    }

    // Update scene after each timer event
    updateScene();
}
